proQuestSpecialTagToFunc = {
    #'DOI' : lambda R : None,
    'authorsShort' : lambda R : R['Author'],
    #'grants' : lambda R : [],
    #'address' : '',
    'selfCitation' : lambda R: R.createCitation(),
    'beginningPage' : lambda R : 0,
    'keywords' : lambda R : R['Identifier / keyword'],
    'abstract' : lambda R : R['Abstract'],
    #'citations' : '',
    'authorsFull' : lambda R : 'Author',
    #'volume' : '',
    'year' : lambda R : R['Degree date'],
    #'j9' : '',
    #'journal' : '',
    'title' : lambda R : R['Title'],
    'id' : lambda R : R.id,
}
