#Written by Reid McIlroy-Young for Dr. John McLevey, University of Waterloo 2015
"""Two functions based on _matplotlib_ for generating nicer looking graphs

This is the only module that depends on anything besides _networkx_, it depends on [numpy](http://www.numpy.org/), [scipy](https://www.scipy.org/) and [matplotlib](http://matplotlib.org/).
"""

from .plotting import graphDensityContourPlot, quickVisual
